#!/usr/bin/env python3

from debian.deb822 import Deb822, Release
import email.utils
import os
import sys
import shutil
import subprocess
import argparse
import time
from datetime import timedelta
from collections import defaultdict
from itertools import product

have_qemu = os.getenv("HAVE_QEMU", "yes") == "yes"
have_binfmt = os.getenv("HAVE_BINFMT", "yes") == "yes"
run_ma_same_tests = os.getenv("RUN_MA_SAME_TESTS", "yes") == "yes"
use_host_apt_config = os.getenv("USE_HOST_APT_CONFIG", "no") == "yes"
cmd = os.getenv("CMD", "./mmdebstrap")

default_dist = os.getenv("DEFAULT_DIST", "unstable")
all_dists = ["oldstable", "stable", "testing", "unstable"]
default_mode = "auto"
all_modes = ["auto", "root", "unshare", "fakechroot", "chrootless"]
default_variant = "apt"
all_variants = [
    "extract",
    "custom",
    "essential",
    "apt",
    "minbase",
    "buildd",
    "-",
    "standard",
]
default_format = "auto"
all_formats = ["auto", "directory", "tar", "squashfs", "ext2", "ext4", "null"]

mirror = os.getenv("mirror", "http://127.0.0.1/debian")
hostarch = subprocess.check_output(["dpkg", "--print-architecture"]).decode().strip()

release_path = f"./shared/cache/debian/dists/{default_dist}/InRelease"
if not os.path.exists(release_path):
    print("path doesn't exist:", release_path, file=sys.stderr)
    print("run ./make_mirror.sh first", file=sys.stderr)
    exit(1)
if os.getenv("SOURCE_DATE_EPOCH") is not None:
    s_d_e = os.getenv("SOURCE_DATE_EPOCH")
else:
    with open(release_path) as f:
        rel = Release(f)
    s_d_e = str(email.utils.mktime_tz(email.utils.parsedate_tz(rel["Date"])))

separator = (
    "------------------------------------------------------------------------------"
)


def skip(condition, dist, mode, variant, fmt):
    if not condition:
        return ""
    for line in condition.splitlines():
        if not line:
            continue
        if eval(line):
            return line.strip()
    return ""


def parse_config(confname):
    config_dict = defaultdict(dict)
    config_order = list()
    all_vals = {
        "Dists": all_dists,
        "Modes": all_modes,
        "Variants": all_variants,
        "Formats": all_formats,
    }
    with open(confname) as f:
        for test in Deb822.iter_paragraphs(f):
            if "Test" not in test.keys():
                print("Test without name", file=sys.stderr)
                exit(1)
            name = test["Test"]
            config_order.append(name)
            for k in test.keys():
                v = test[k]
                if k not in [
                    "Test",
                    "Dists",
                    "Modes",
                    "Variants",
                    "Formats",
                    "Skip-If",
                    "Needs-QEMU",
                    "Needs-Root",
                    "Needs-APT-Config",
                ]:
                    print(f"Unknown field name {k} in test {name}")
                    exit(1)
                if k in all_vals.keys():
                    if v == "default":
                        print(
                            f"Setting {k} to default in Test {name} is redundant",
                            file=sys.stderr,
                        )
                        exit(1)
                    if v == "any":
                        v = all_vals[k]
                    else:
                        # else, split the value by whitespace
                        v = v.split()
                        for i in v:
                            if i not in all_vals[k]:
                                print(
                                    f"{i} is not a valid value for {k}", file=sys.stderr
                                )
                                exit(1)
                config_dict[name][k] = v
    return config_order, config_dict


def format_test(num, total, name, dist, mode, variant, fmt, config_dict):
    ret = f"({num}/{total}) {name}"
    if len(config_dict[name].get("Dists", [])) > 1:
        ret += f" --dist={dist}"
    if len(config_dict[name].get("Modes", [])) > 1:
        ret += f" --mode={mode}"
    if len(config_dict[name].get("Variants", [])) > 1:
        ret += f" --variant={variant}"
    if len(config_dict[name].get("Formats", [])) > 1:
        ret += f" --format={fmt}"
    return ret


def print_time_per_test(time_per_test, name="test"):
    print(
        f"average time per {name}:",
        sum(time_per_test.values(), start=timedelta()) / len(time_per_test),
        file=sys.stderr,
    )
    print(
        f"median time per {name}:",
        sorted(time_per_test.values())[len(time_per_test) // 2],
        file=sys.stderr,
    )
    head_tail_num = 10
    print(f"{head_tail_num} fastests {name}s:", file=sys.stderr)
    for k, v in sorted(time_per_test.items(), key=lambda i: i[1])[
        : min(head_tail_num, len(time_per_test))
    ]:
        print(f"    {k}: {v}", file=sys.stderr)
    print(f"{head_tail_num} slowest {name}s:", file=sys.stderr)
    for k, v in sorted(time_per_test.items(), key=lambda i: i[1], reverse=True)[
        : min(head_tail_num, len(time_per_test))
    ]:
        print(f"    {k}: {v}", file=sys.stderr)


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("test", nargs="*", help="only run these tests")
    parser.add_argument(
        "-x",
        "--exitfirst",
        action="store_const",
        dest="maxfail",
        const=1,
        help="exit instantly on first error or failed test.",
    )
    parser.add_argument(
        "--maxfail",
        metavar="num",
        action="store",
        type=int,
        dest="maxfail",
        default=0,
        help="exit after first num failures or errors.",
    )
    parser.add_argument(
        "--mode",
        metavar="mode",
        help=f"only run tests with this mode (Default = {default_mode})",
    )
    parser.add_argument(
        "--dist",
        metavar="dist",
        help=f"only run tests with this dist (Default = {default_dist})",
    )
    parser.add_argument(
        "--variant",
        metavar="variant",
        help=f"only run tests with this variant (Default = {default_variant})",
    )
    parser.add_argument(
        "--format",
        metavar="format",
        help=f"only run tests with this format (Default = {default_format})",
    )
    parser.add_argument(
        "--skip", metavar="test", action="append", help="skip this test"
    )
    args = parser.parse_args()

    # copy over files from git or as distributed
    for git, dist, target in [
        ("./mmdebstrap", "/usr/bin/mmdebstrap", "mmdebstrap"),
        ("./tarfilter", "/usr/bin/mmtarfilter", "tarfilter"),
        (
            "./proxysolver",
            "/usr/lib/apt/solvers/mmdebstrap-dump-solution",
            "proxysolver",
        ),
        (
            "./ldconfig.fakechroot",
            "/usr/libexec/mmdebstrap/ldconfig.fakechroot",
            "ldconfig.fakechroot",
        ),
    ]:
        if os.path.exists(git):
            shutil.copy(git, f"shared/{target}")
        else:
            shutil.copy(dist, f"shared/{target}")
    # copy over hooks from git or as distributed
    if os.path.exists("hooks"):
        shutil.copytree("hooks", "shared/hooks", dirs_exist_ok=True)
    else:
        shutil.copytree(
            "/usr/share/mmdebstrap/hooks", "shared/hooks", dirs_exist_ok=True
        )

    # parse coverage.txt
    config_order, config_dict = parse_config("coverage.txt")

    indirbutnotcovered = set(
        [d for d in os.listdir("tests") if not d.startswith(".")]
    ) - set(config_order)
    if indirbutnotcovered:
        print(
            "test(s) missing from coverage.txt: %s"
            % (", ".join(sorted(indirbutnotcovered))),
            file=sys.stderr,
        )
        exit(1)
    coveredbutnotindir = set(config_order) - set(
        [d for d in os.listdir("tests") if not d.startswith(".")]
    )
    if coveredbutnotindir:
        print(
            "test(s) missing from ./tests: %s"
            % (", ".join(sorted(coveredbutnotindir))),
            file=sys.stderr,
        )

        exit(1)

    # produce the list of tests using the cartesian product of all allowed
    # dists, modes, variants and formats of a given test
    tests = []
    for name in config_order:
        test = config_dict[name]
        for dist, mode, variant, fmt in product(
            test.get("Dists", [default_dist]),
            test.get("Modes", [default_mode]),
            test.get("Variants", [default_variant]),
            test.get("Formats", [default_format]),
        ):
            skipreason = skip(test.get("Skip-If"), dist, mode, variant, fmt)
            if skipreason:
                tt = ("skip", skipreason)
            elif (
                test.get("Needs-APT-Config", "false") == "true" and use_host_apt_config
            ):
                tt = ("skip", "test cannot use host apt config")
            elif have_qemu:
                tt = "qemu"
            elif test.get("Needs-QEMU", "false") == "true":
                tt = ("skip", "test needs QEMU")
            elif test.get("Needs-Root", "false") == "true":
                tt = "sudo"
            elif mode == "root":
                tt = "sudo"
            else:
                tt = "null"
            tests.append((tt, name, dist, mode, variant, fmt))

    torun = []
    num_tests = len(tests)
    if args.test:
        # check if all given tests are either a valid name or a valid number
        for test in args.test:
            if test in [name for (_, name, _, _, _, _) in tests]:
                continue
            if not test.isdigit():
                print(f"cannot find test named {test}", file=sys.stderr)
                exit(1)
            if int(test) >= len(tests) or int(test) <= 0 or str(int(test)) != test:
                print(f"test number {test} doesn't exist", file=sys.stderr)
                exit(1)

        for i, (_, name, _, _, _, _) in enumerate(tests):
            # if either the number or test name matches, then we use this test,
            # otherwise we skip it
            if name in args.test:
                torun.append(i)
            if str(i + 1) in args.test:
                torun.append(i)
        num_tests = len(torun)

    starttime = time.time()
    skipped = defaultdict(list)
    failed = []
    num_success = 0
    num_finished = 0
    time_per_test = {}
    acc_time_per_test = defaultdict(list)
    for i, (test, name, dist, mode, variant, fmt) in enumerate(tests):
        if torun and i not in torun:
            continue
        print(separator, file=sys.stderr)
        print("(%d/%d) %s" % (i + 1, len(tests), name), file=sys.stderr)
        print("dist: %s" % dist, file=sys.stderr)
        print("mode: %s" % mode, file=sys.stderr)
        print("variant: %s" % variant, file=sys.stderr)
        print("format: %s" % fmt, file=sys.stderr)
        if num_finished > 0:
            currenttime = time.time()
            timeleft = timedelta(
                seconds=int(
                    (num_tests - num_finished)
                    * (currenttime - starttime)
                    / num_finished
                )
            )
            print("time left: %s" % timeleft, file=sys.stderr)
        if failed:
            print("failed: %d" % len(failed), file=sys.stderr)
        num_finished += 1
        with open("tests/" + name) as fin, open("shared/test.sh", "w") as fout:
            for line in fin:
                line = line.replace("{{ CMD }}", cmd)
                line = line.replace("{{ SOURCE_DATE_EPOCH }}", s_d_e)
                line = line.replace("{{ DIST }}", dist)
                line = line.replace("{{ MIRROR }}", mirror)
                line = line.replace("{{ MODE }}", mode)
                line = line.replace("{{ VARIANT }}", variant)
                line = line.replace("{{ FORMAT }}", fmt)
                line = line.replace("{{ HOSTARCH }}", hostarch)
                fout.write(line)
        # ignore:
        # SC2016 Expressions don't expand in single quotes, use double quotes for that.
        # SC2050 This expression is constant. Did you forget the $ on a variable?
        # SC2194 This word is constant. Did you forget the $ on a variable?
        shellcheck = subprocess.run(
            [
                "shellcheck",
                "--exclude=SC2050,SC2194,SC2016",
                "-f",
                "gcc",
                "shared/test.sh",
            ],
            check=False,
            stdout=subprocess.PIPE,
        ).stdout.decode()
        shfmt = subprocess.run(
            [
                "shfmt",
                "--posix",
                "--binary-next-line",
                "--case-indent",
                "--indent",
                "2",
                "--simplify",
                "-d",
                "shared/test.sh",
            ],
            check=False,
            stdout=subprocess.PIPE,
        ).stdout.decode()
        argv = None
        match test:
            case "qemu":
                argv = ["./run_qemu.sh"]
            case "sudo":
                argv = ["./run_null.sh", "SUDO"]
            case "null":
                argv = ["./run_null.sh"]
            case ("skip", reason):
                skipped[reason].append(
                    format_test(
                        i + 1, len(tests), name, dist, mode, variant, fmt, config_dict
                    )
                )
                print(f"skipped because of {reason}", file=sys.stderr)
                continue
        print(separator, file=sys.stderr)
        if args.skip and name in args.skip:
            print(f"skipping because of --skip={name}", file=sys.stderr)
            continue
        if args.dist and args.dist != dist:
            print(f"skipping because of --dist={args.dist}", file=sys.stderr)
            continue
        if args.mode and args.mode != mode:
            print(f"skipping because of --mode={args.mode}", file=sys.stderr)
            continue
        if args.variant and args.variant != variant:
            print(f"skipping because of --variant={args.variant}", file=sys.stderr)
            continue
        if args.format and args.format != fmt:
            print(f"skipping because of --format={args.format}", file=sys.stderr)
            continue
        before = time.time()
        proc = subprocess.Popen(argv)
        try:
            proc.wait()
        except KeyboardInterrupt:
            proc.terminate()
            proc.wait()
            break
        after = time.time()
        walltime = timedelta(seconds=int(after - before))
        formated_test_name = format_test(
            i + 1, len(tests), name, dist, mode, variant, fmt, config_dict
        )
        time_per_test[formated_test_name] = walltime
        acc_time_per_test[name].append(walltime)
        print(separator, file=sys.stderr)
        print(f"duration: {walltime}", file=sys.stderr)
        if proc.returncode != 0 or shellcheck != "" or shfmt != "":
            if shellcheck != "":
                print(shellcheck)
            if shfmt != "":
                print(shfmt)
            failed.append(formated_test_name)
            print("result: FAILURE", file=sys.stderr)
        else:
            print("result: SUCCESS", file=sys.stderr)
            num_success += 1
        if args.maxfail and len(failed) >= args.maxfail:
            break
    print(separator, file=sys.stderr)
    print(
        "successfully ran %d tests" % num_success,
        file=sys.stderr,
    )
    if skipped:
        print("skipped %d:" % sum([len(v) for v in skipped.values()]), file=sys.stderr)
        for reason, l in skipped.items():
            print(f"skipped because of {reason}:", file=sys.stderr)
            for t in l:
                print(f"    {t}", file=sys.stderr)
    if len(time_per_test) > 1:
        print_time_per_test(time_per_test)
    if len(acc_time_per_test) > 1:
        print_time_per_test(
            {
                f"{len(v)}x {k}": sum(v, start=timedelta())
                for k, v in acc_time_per_test.items()
            },
            "accumulated test",
        )
    if failed:
        print("failed %d:" % len(failed), file=sys.stderr)
        for f in failed:
            print(f, file=sys.stderr)
    currenttime = time.time()
    walltime = timedelta(seconds=int(currenttime - starttime))
    print(f"total runtime: {walltime}", file=sys.stderr)
    if failed:
        exit(1)


if __name__ == "__main__":
    main()



#!/usr/bin/python3
import argparse
import pathlib
import subprocess
import tempfile
import pathlib

__author__ = "Trent W. Buck"
__copyright__ = "Copyright © 2020 Trent W. Buck"
__license__ = "expat"

__doc__ = """ build the simplest Debian Live image that can boot

This uses mmdebstrap to do the heavy lifting;
it can run entirely without root privileges.
It emits a USB key disk image that contains a bootable EFI ESP,
which in turn includes a bootloader (refind), kernel, ramdisk, and filesystem.squashfs.

NOTE: this is the simplest config possible.
      It lacks CRITICAL SECURITY AND DATA LOSS packages, such as amd64-microcode and smartd.
"""

parser = argparse.ArgumentParser(description=__doc__)
parser.add_argument(
    "output_file", nargs="?", default=pathlib.Path("filesystem.img"), type=pathlib.Path
)
args = parser.parse_args()


filesystem_img_size = "256M"  # big enough to include filesystem.squashfs + about 64M of bootloader, kernel, and ramdisk.
esp_offset = 1024 * 1024  # 1MiB
esp_label = "UEFI-ESP"  # max 8 bytes for FAT32
live_media_path = "debian-live"

with tempfile.TemporaryDirectory(prefix="debian-live-bullseye-amd64-minimal.") as td:
    td = pathlib.Path(td)
    subprocess.check_call(
        [
            "mmdebstrap",
            "--mode=unshare",
            "--variant=apt",
            '--aptopt=Acquire::http::Proxy "http://apt-cacher-ng.cyber.com.au:3142"',
            '--aptopt=Acquire::https::Proxy "DIRECT"',
            "--dpkgopt=force-unsafe-io",
            "--include=linux-image-amd64 init initramfs-tools live-boot netbase",
            "--include=dbus",  # https://bugs.debian.org/814758
            "--include=live-config iproute2 keyboard-configuration locales sudo user-setup",
            "--include=ifupdown isc-dhcp-client",  # live-config doesn't support systemd-networkd yet.
            # Do the **BARE MINIMUM** to make a USB key that can boot on X86_64 UEFI.
            # We use mtools so we do not ever need root privileges.
            # We can't use mkfs.vfat, as that needs kpartx or losetup (i.e. root).
            # We can't use mkfs.udf, as that needs mount (i.e. root).
            # We can't use "refind-install --usedefault" as that runs mount(8) (i.e. root).
            # We don't use genisoimage because
            # 1) ISO9660 must die;
            # 2) incomplete UDF 1.5+ support;
            # 3) resulting filesystem can't be tweaked after flashing (e.g. debian-live/site.dir/etc/systemd/network/up.network).
            #
            # We use refind because 1) I hate grub; and 2) I like refind.
            # If you want aarch64 or ia32 you need to install their BOOTxxx.EFI files.
            # If you want kernel+initrd on something other than FAT, you need refind/drivers_xxx/xxx_xxx.EFI.
            #
            # FIXME: with qemu in UEFI mode (OVMF), I get dumped into startup.nsh (UEFI REPL).
            #        From there, I can manually type in "FS0:\EFI\BOOT\BOOTX64.EFI" to start refind, tho.
            #        So WTF is its problem?  Does it not support fallback bootloader?
            "--include=refind parted mtools",
            "--essential-hook=echo refind refind/install_to_esp boolean false | chroot $1 debconf-set-selections",
            "--customize-hook=echo refind refind/install_to_esp boolean true  | chroot $1 debconf-set-selections",
            "--customize-hook=chroot $1 mkdir -p /boot/USB /boot/EFI/BOOT",
            "--customize-hook=chroot $1 cp /usr/share/refind/refind/refind_x64.efi /boot/EFI/BOOT/BOOTX64.EFI",
            f"--customize-hook=chroot $1 truncate --size={filesystem_img_size} /boot/USB/filesystem.img",
            f"--customize-hook=chroot $1 parted --script --align=optimal /boot/USB/filesystem.img  mklabel gpt  mkpart {esp_label} {esp_offset}b 100%  set 1 esp on",
            f"--customize-hook=chroot $1 mformat -i /boot/USB/filesystem.img@@{esp_offset} -F -v {esp_label}",
            f"--customize-hook=chroot $1 mmd     -i /boot/USB/filesystem.img@@{esp_offset} ::{live_media_path}",
            f"""--customize-hook=echo '"Boot with default options" "boot=live live-media-path={live_media_path}"' >$1/boot/refind_linux.conf""",
            # NOTE: find sidesteps the "glob expands before chroot applies" problem.
            f"""--customize-hook=chroot $1 find -O3 /boot/ -xdev -mindepth 1 -maxdepth 1 -regextype posix-egrep -iregex '.*/(EFI|refind_linux.conf|vmlinuz.*|initrd.img.*)' -exec mcopy -vsbpm -i /boot/USB/filesystem.img@@{esp_offset} {{}} :: ';'""",
            # FIXME: copy-out doesn't handle sparseness, so is REALLY slow (about 50 seconds).
            # Therefore instead leave it in the squashfs, and extract it later.
            #  f'--customize-hook=copy-out /boot/USB/filesystem.img /tmp/',
            #  f'--customize-hook=chroot $1 rm /boot/USB/filesystem.img',
            "bullseye",
            td / "filesystem.squashfs",
        ]
    )

    with args.output_file.open("wb") as f:
        subprocess.check_call(
            ["rdsquashfs", "--cat=boot/USB/filesystem.img", td / "filesystem.squashfs"],
            stdout=f,
        )
    subprocess.check_call(
        [
            "mcopy",
            "-i",
            f"{args.output_file}@@{esp_offset}",
            td / "filesystem.squashfs",
            f"::{live_media_path}/filesystem.squashfs",
        ]
    )



#!/usr/bin/env python3

import sys
import os
import time
import http.client
import http.server
from io import StringIO
import pathlib
import urllib.parse

oldcachedir = None
newcachedir = None
readonly = False


class ProxyRequestHandler(http.server.BaseHTTPRequestHandler):
    def do_GET(self):
        assert int(self.headers.get("Content-Length", 0)) == 0
        assert self.headers["Host"]
        pathprefix = "http://" + self.headers["Host"] + "/"
        assert self.path.startswith(pathprefix)
        sanitizedpath = urllib.parse.unquote(self.path.removeprefix(pathprefix))
        oldpath = oldcachedir / sanitizedpath
        newpath = newcachedir / sanitizedpath

        if not readonly:
            newpath.parent.mkdir(parents=True, exist_ok=True)

        # just send back to client
        if newpath.exists():
            print(f"proxy cached: {self.path}", file=sys.stderr)
            self.wfile.write(b"HTTP/1.1 200 OK\r\n")
            self.send_header("Content-Length", newpath.stat().st_size)
            self.end_headers()
            with newpath.open(mode="rb") as new:
                while True:
                    buf = new.read(64 * 1024)  # same as shutil uses
                    if not buf:
                        break
                    self.wfile.write(buf)
            self.wfile.flush()
            return

        if readonly:
            newpath = pathlib.Path("/dev/null")

        # copy from oldpath to newpath and send back to client
        # Only take files from the old cache if they are .deb files or Packages
        # files in the by-hash directory as only those are unique by their path
        # name. Other files like InRelease files have to be downloaded afresh.
        if oldpath.exists() and (
            oldpath.suffix == ".deb" or "by-hash" in oldpath.parts
        ):
            print(f"proxy cached: {self.path}", file=sys.stderr)
            self.wfile.write(b"HTTP/1.1 200 OK\r\n")
            self.send_header("Content-Length", oldpath.stat().st_size)
            self.end_headers()
            with oldpath.open(mode="rb") as old, newpath.open(mode="wb") as new:
                # we are not using shutil.copyfileobj() because we want to
                # write to two file objects simultaneously
                while True:
                    buf = old.read(64 * 1024)  # same as shutil uses
                    if not buf:
                        break
                    self.wfile.write(buf)
                    new.write(buf)
            self.wfile.flush()
            return

        # download fresh copy
        try:
            print(f"\rproxy download: {self.path}", file=sys.stderr)
            conn = http.client.HTTPConnection(self.headers["Host"], timeout=5)
            conn.request("GET", self.path, None, dict(self.headers))
            res = conn.getresponse()
            assert (res.status, res.reason) == (200, "OK"), (res.status, res.reason)
            self.wfile.write(b"HTTP/1.1 200 OK\r\n")
            for k, v in res.getheaders():
                # do not allow a persistent connection
                if k == "connection":
                    continue
                self.send_header(k, v)
            self.end_headers()
            with newpath.open(mode="wb") as f:
                # we are not using shutil.copyfileobj() because we want to
                # write to two file objects simultaneously and throttle the
                # writing speed to 1024 kB/s
                while True:
                    buf = res.read(64 * 1024)  # same as shutil uses
                    if not buf:
                        break
                    self.wfile.write(buf)
                    f.write(buf)
                    time.sleep(64 / 1024)  # 1024 kB/s
            self.wfile.flush()
        except Exception as e:
            self.send_error(502)


def main():
    global oldcachedir, newcachedir, readonly
    if sys.argv[1] == "--readonly":
        readonly = True
        oldcachedir = pathlib.Path(sys.argv[2])
        newcachedir = pathlib.Path(sys.argv[3])
    else:
        oldcachedir = pathlib.Path(sys.argv[1])
        newcachedir = pathlib.Path(sys.argv[2])
    print(f"starting caching proxy for {newcachedir}", file=sys.stderr)
    httpd = http.server.ThreadingHTTPServer(
        server_address=("", 8080), RequestHandlerClass=ProxyRequestHandler
    )
    httpd.serve_forever()


if __name__ == "__main__":
    main()



#!/usr/bin/python3
import argparse
import pathlib
import subprocess
import tempfile
import pathlib

__author__ = "Trent W. Buck"
__copyright__ = "Copyright © 2020 Trent W. Buck"
__license__ = "expat"

__doc__ = """ build a Debian Live image that can install Debian 11 on ZFS 2

This uses mmdebstrap to do the heavy lifting;
it can run entirely without root privileges.
It emits a USB key disk image that contains a bootable EFI ESP,
which in turn includes a bootloader (refind), kernel, ramdisk, and filesystem.squashfs.

"""

parser = argparse.ArgumentParser(description=__doc__)
parser.add_argument(
    "output_file", nargs="?", default=pathlib.Path("filesystem.img"), type=pathlib.Path
)
parser.add_argument(
    "--timezone",
    default="Australia/Melbourne",
    type=lambda s: s.split("/"),
    help='NOTE: MUST be "Area/Zone" not e.g. "UTC", for now',
)
parser.add_argument(
    "--locale", default="en_AU.UTF-8", help='NOTE: MUST end in ".UTF-8", for now'
)
args = parser.parse_args()


filesystem_img_size = "512M"  # big enough to include filesystem.squashfs + about 64M of bootloader, kernel, and ramdisk.
esp_offset = 1024 * 1024  # 1MiB
esp_label = "UEFI-ESP"  # max 8 bytes for FAT32
live_media_path = "debian-live"

with tempfile.TemporaryDirectory(prefix="debian-sid-zfs.") as td:
    td = pathlib.Path(td)
    subprocess.check_call(
        [
            "mmdebstrap",
            "--mode=unshare",
            "--variant=apt",
            '--aptopt=Acquire::http::Proxy "http://apt-cacher-ng.cyber.com.au:3142"',
            '--aptopt=Acquire::https::Proxy "DIRECT"',
            "--dpkgopt=force-unsafe-io",
            "--components=main contrib non-free",  # needed for CPU security patches
            "--include=init initramfs-tools xz-utils live-boot netbase",
            "--include=dbus",  # https://bugs.debian.org/814758
            "--include=linux-image-amd64 firmware-linux",
            # Have ZFS 2.0 support.
            "--include=zfs-dkms zfsutils-linux zfs-zed build-essential linux-headers-amd64",  # ZFS 2 support
            # Make the initrd a little smaller (41MB -> 20MB), at the expensive of significantly slower image build time.
            "--include=zstd",
            "--essential-hook=mkdir -p $1/etc/initramfs-tools/conf.d",
            "--essential-hook=>$1/etc/initramfs-tools/conf.d/zstd echo COMPRESS=zstd",
            # Be the equivalent of Debian Live GNOME
            # '--include=live-task-gnome',
            #'--include=live-task-xfce',
            # FIXME: enable this?  It makes live-task-xfce go from 1G to 16G... so no.
            #'--aptopt=Apt::Install-Recommends "true"',
            # ...cherry-pick instead
            # UPDATE: debian-installer-launcher DOES NOT WORK because we don't load crap SPECIFICALLY into /live/installer, in the ESP.
            # UPDATE: network-manager-gnome DOES NOT WORK, nor is systemd-networkd auto-started... WTF?
            #         end result is no networking.
            #'--include=live-config user-setup sudo firmware-linux haveged',
            #'--include=calamares-settings-debian udisks2',  # 300MB weirdo Qt GUI debian installer
            #'--include=xfce4-terminal',
            # x86_64 CPUs are undocumented proprietary RISC chips that EMULATE a documented x86_64 CISC ISA.
            # The emulator is called "microcode", and is full of security vulnerabilities.
            # Make sure security patches for microcode for *ALL* CPUs are included.
            # By default, it tries to auto-detect the running CPU, so only patches the CPU of the build server.
            "--include=intel-microcode amd64-microcode iucode-tool",
            "--essential-hook=>$1/etc/default/intel-microcode echo IUCODE_TOOL_INITRAMFS=yes IUCODE_TOOL_SCANCPUS=no",
            "--essential-hook=>$1/etc/default/amd64-microcode echo AMD64UCODE_INITRAMFS=yes",
            "--dpkgopt=force-confold",  # Work around https://bugs.debian.org/981004
            # DHCP/DNS/SNTP clients...
            # FIXME: use live-config ?
            "--include=libnss-resolve libnss-myhostname systemd-timesyncd",
            "--customize-hook=chroot $1 cp -alf /lib/systemd/resolv.conf /etc/resolv.conf",  # This probably needs to happen LAST
            # FIXME: fix resolv.conf to point to resolved, not "copy from the build-time OS"
            # FIXME: fix hostname & hosts to not exist, not "copy from the build-time OS"
            "--customize-hook=systemctl --root=$1 enable systemd-networkd systemd-timesyncd",  # is this needed?
            # Run a DHCP client on *ALL* ifaces.
            # Consider network "up" (start sshd and local login prompt) when *ANY* (not ALL) ifaces are up.
            "--customize-hook=>$1/etc/systemd/network/up.network printf '%s\n'  '[Match]' Name='en*' '[Network]' DHCP=yes",  # try DHCP on all ethernet ifaces
            "--customize-hook=mkdir $1/etc/systemd/system/systemd-networkd-wait-online.service.d",
            "--customize-hook=>$1/etc/systemd/system/systemd-networkd-wait-online.service.d/any-not-all.conf printf '%s\n' '[Service]' 'ExecStart=' 'ExecStart=/lib/systemd/systemd-networkd-wait-online --any'",
            # Hope there's a central smarthost SMTP server called "mail" in the local search domain.
            # FIXME: can live-config do this?
            "--include=msmtp-mta",
            "--customize-hook=>$1/etc/msmtprc printf '%s\n' 'account default' 'syslog LOG_MAIL' 'host mail' 'auto_from on'",
            # Hope there's a central RELP logserver called "logserv" in the local domain.
            # FIXME: can live-config do this?
            "--include=rsyslog-relp",
            """--customize-hook=>$1/etc/rsyslog.conf printf '%s\n' 'module(load="imuxsock")' 'module(load="imklog")' 'module(load="omrelp")' 'action(type="omrelp" target="logserv" port="2514" template="RSYSLOG_SyslogProtocol23Format")'""",
            # Run self-tests on all discoverable hard disks, and (try to) email if something goes wrong.
            "--include=smartmontools bsd-mailx",
            "--customize-hook=>$1/etc/smartd.conf echo 'DEVICESCAN -n standby,15 -a -o on -S on -s (S/../../7/00|L/../01/./01) -t -H -m root -M once'",
            # For rarely-updated, rarely-rebooted SOEs, apply what security updates we can into transient tmpfs COW.
            # This CANNOT apply kernel security updates (though it will download them).
            # This CANNOT make the upgrades persistent across reboots (they re-download each boot).
            # FIXME: Would it be cleaner to set Environment=NEEDRESTART_MODE=a in
            #        apt-daily-upgrade.service and/or
            #        unattended-upgrades.service, so
            #        needrestart is noninteractive only when apt is noninteractive?
            "--include=unattended-upgrades needrestart",
            "--customize-hook=echo 'unattended-upgrades unattended-upgrades/enable_auto_updates boolean true' | chroot $1 debconf-set-selections",
            """--customize-hook=>$1/etc/needrestart/conf.d/unattended-needrestart.conf  echo '$nrconf{restart} = "a";'""",  # https://bugs.debian.org/894444
            # Do an apt update & apt upgrade at boot time (as well as @daily).
            # The lack of /etc/machine-id causes these to be implicitly enabled.
            # FIXME: use dropin in /etc.
            "--customize-hook=>>$1/lib/systemd/system/apt-daily.service		printf '%s\n' '[Install]' 'WantedBy=multi-user.target'",
            "--customize-hook=>>$1/lib/systemd/system/apt-daily-upgrade.service	printf '%s\n' '[Install]' 'WantedBy=multi-user.target'",
            # FIXME: add support for this stuff (for the non-live final install this happens via ansible):
            #
            #            unattended-upgrades
            #            smartd
            #            networkd (boot off ANY NIC, not EVERY NIC -- https://github.com/systemd/systemd/issues/9714)
            #            refind (bootloader config)
            #            misc safety nets
            #            double-check that mmdebstrap's machine-id support works properly
            # Bare minimum to let me SSH in.
            # FIXME: make this configurable.
            # FIXME: trust a CA certificate instead -- see Zero Trust SSH, Jeremy Stott, LCA 2020 <https://youtu.be/lYzklWPTbsQ>
            # WARNING: tinysshd does not support RSA, nor MaxStartups, nor sftp (unless you also install openssh-client, which is huge).
            # FIXME: double-check no host keys are baked into the image (openssh-server and dropbear do this).
            "--include=tinysshd rsync",
            "--essential-hook=install -dm700 $1/root/.ssh",
            '--essential-hook=echo "ssh-ed25519 AAAAC3NzaC1lZDI1NTE5AAAAIIapAZ0E0353DaY6xBnasvu/DOvdWdKQ6RQURwq4l6Wu twb@cyber.com.au (Trent W. Buck)" >$1/root/.ssh/authorized_keys',
            # Bare minimum to let me log in locally.
            # DO NOT use this on production builds!
            "--essential-hook=chroot $1 passwd --delete root",
            # Configure language (not needed to boot).
            # Racism saves a **LOT** of space -- something like 2GB for Debian Live images.
            # FIXME: use live-config instead?
            "--include=locales localepurge",
            f"--essential-hook=echo locales locales/default_environment_locale   select {args.locale}       | chroot $1 debconf-set-selections",
            f"--essential-hook=echo locales locales/locales_to_be_generated multiselect {args.locale} UTF-8 | chroot $1 debconf-set-selections",
            # FIXME: https://bugs.debian.org/603700
            "--customize-hook=chroot $1 sed -i /etc/locale.nopurge -e 's/^USE_DPKG/#ARGH#&/'",
            "--customize-hook=chroot $1 localepurge",
            "--customize-hook=chroot $1 sed -i /etc/locale.nopurge -e 's/^#ARGH#//'",
            # Removing documentation also saves a LOT of space.
            "--dpkgopt=path-exclude=/usr/share/doc/*",
            "--dpkgopt=path-exclude=/usr/share/info/*",
            "--dpkgopt=path-exclude=/usr/share/man/*",
            "--dpkgopt=path-exclude=/usr/share/omf/*",
            "--dpkgopt=path-exclude=/usr/share/help/*",
            "--dpkgopt=path-exclude=/usr/share/gnome/help/*",
            # Configure timezone (not needed to boot)`
            # FIXME: use live-config instead?
            "--include=tzdata",
            f"--essential-hook=echo tzdata tzdata/Areas                    select {args.timezone[0]} | chroot $1 debconf-set-selections",
            f"--essential-hook=echo tzdata tzdata/Zones/{args.timezone[0]} select {args.timezone[1]} | chroot $1 debconf-set-selections",
            # Do the **BARE MINIMUM** to make a USB key that can boot on X86_64 UEFI.
            # We use mtools so we do not ever need root privileges.
            # We can't use mkfs.vfat, as that needs kpartx or losetup (i.e. root).
            # We can't use mkfs.udf, as that needs mount (i.e. root).
            # We can't use "refind-install --usedefault" as that runs mount(8) (i.e. root).
            # We don't use genisoimage because
            # 1) ISO9660 must die;
            # 2) incomplete UDF 1.5+ support;
            # 3) resulting filesystem can't be tweaked after flashing (e.g. debian-live/site.dir/etc/systemd/network/up.network).
            #
            # We use refind because 1) I hate grub; and 2) I like refind.
            # If you want aarch64 or ia32 you need to install their BOOTxxx.EFI files.
            # If you want kernel+initrd on something other than FAT, you need refind/drivers_xxx/xxx_xxx.EFI.
            #
            # FIXME: with qemu in UEFI mode (OVMF), I get dumped into startup.nsh (UEFI REPL).
            #        From there, I can manually type in "FS0:\EFI\BOOT\BOOTX64.EFI" to start refind, tho.
            #        So WTF is its problem?  Does it not support fallback bootloader?
            "--include=refind parted mtools",
            "--essential-hook=echo refind refind/install_to_esp boolean false | chroot $1 debconf-set-selections",
            "--customize-hook=echo refind refind/install_to_esp boolean true  | chroot $1 debconf-set-selections",
            "--customize-hook=chroot $1 mkdir -p /boot/USB /boot/EFI/BOOT",
            "--customize-hook=chroot $1 cp /usr/share/refind/refind/refind_x64.efi /boot/EFI/BOOT/BOOTX64.EFI",
            "--customize-hook=chroot $1 cp /usr/share/refind/refind/refind.conf-sample /boot/EFI/BOOT/refind.conf",
            f"--customize-hook=chroot $1 truncate --size={filesystem_img_size} /boot/USB/filesystem.img",
            f"--customize-hook=chroot $1 parted --script --align=optimal /boot/USB/filesystem.img  mklabel gpt  mkpart {esp_label} {esp_offset}b 100%  set 1 esp on",
            f"--customize-hook=chroot $1 mformat -i /boot/USB/filesystem.img@@{esp_offset} -F -v {esp_label}",
            f"--customize-hook=chroot $1 mmd     -i /boot/USB/filesystem.img@@{esp_offset} ::{live_media_path}",
            f"""--customize-hook=echo '"Boot with default options" "boot=live live-media-path={live_media_path}"' >$1/boot/refind_linux.conf""",
            f"""--customize-hook=chroot $1 find /boot/ -xdev -mindepth 1 -maxdepth 1 -not -name filesystem.img -not -name USB -exec mcopy -vsbpm -i /boot/USB/filesystem.img@@{esp_offset} {{}} :: ';'""",
            # FIXME: copy-out doesn't handle sparseness, so is REALLY slow (about 50 seconds).
            # Therefore instead leave it in the squashfs, and extract it later.
            #  f'--customize-hook=copy-out /boot/USB/filesystem.img /tmp/',
            #  f'--customize-hook=chroot $1 rm /boot/USB/filesystem.img',
            "sid",
            td / "filesystem.squashfs",
        ]
    )

    with args.output_file.open("wb") as f:
        subprocess.check_call(
            ["rdsquashfs", "--cat=boot/USB/filesystem.img", td / "filesystem.squashfs"],
            stdout=f,
        )
    subprocess.check_call(
        [
            "mcopy",
            "-i",
            f"{args.output_file}@@{esp_offset}",
            td / "filesystem.squashfs",
            f"::{live_media_path}/filesystem.squashfs",
        ]
    )
